/*
 * Copyright 2011 Waki Takeya
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.develop.gae.command;

import java.io.IOException;
import java.util.ConcurrentModificationException;

import javax.persistence.OptimisticLockException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import jp.develop.common.util.AmfUtil;
import jp.develop.gae.command.exception.ApplicationException;
import jp.develop.gae.command.exception.AuthenticationException;

/**
 * Command Executer Servlet using AMF48.
 * @author Waki Takeya
 */
public class CommandServlet extends HttpServlet {

	private static final long serialVersionUID = 5886025571535635750L;

	@Override
	protected void doPost(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {

		byte[] result = null;
		boolean appPhase = false;
		try {
			// Decode Parameter.
			ICommand<?> command = AmfUtil.decode(req.getInputStream(), ICommand.class);

			// Command Execution.
			appPhase = true;
			Object data = command.execute();
			appPhase = false;

			// Result.
			result = AmfUtil.encode(new Result(data, Result.SUCCESS, null));
		} catch (AuthenticationException e) {
			result = getAuthenticationErrorResult();
		} catch (ConcurrentModificationException e) {
			result = getOptimisticLockErrorResult();
		} catch (OptimisticLockException e) {
			result = getOptimisticLockFatalResult();
		} catch (ApplicationException e) {
			result = AmfUtil.encode(new Result(null, e.getCode(), e.getMessage()));
		} catch (Throwable e) {
			e.printStackTrace(System.err);
			if (appPhase) {
				result = getApplicationErrorResult();
			} else {
				result = getInternalErrorResult();
			}
		} finally {
			resp.setContentType("application/x-amf");
			resp.setContentLength(result.length);
			resp.getOutputStream().write(result);
		}
	}

	private volatile byte[] authenticationError;
	private byte[] getAuthenticationErrorResult() {
		if (this.authenticationError == null) {
			this.authenticationError = AmfUtil.encode(
					new Result(null, Result.FATAL_AUTHENTICATION, "Authentication Error."));
		}
		return this.authenticationError;
	}

	private volatile byte[] optimisticLockError;
	private byte[] getOptimisticLockErrorResult() {
		if (this.optimisticLockError == null) {
			this.optimisticLockError = AmfUtil.encode(
					new Result(null, Result.ERROR_OPTIMISTIC_LOCK, "Optimistic Lock Error."));
		}
		return this.optimisticLockError;
	}

	private volatile byte[] optimisticLockFatal;
	private byte[] getOptimisticLockFatalResult() {
		if (this.optimisticLockFatal == null) {
			this.optimisticLockFatal = AmfUtil.encode(
					new Result(null, Result.FATAL_OPTIMISTIC_LOCK, "Optimistic Lock Fatal Error."));
		}
		return this.optimisticLockFatal;
	}

	private volatile byte[] applicationError;
	private byte[] getApplicationErrorResult() {
		if (this.applicationError == null) {
			this.applicationError = AmfUtil.encode(
					new Result(null, Result.ERROR_APPLICATION, "Unknown Application Error."));
		}
		return this.applicationError;
	}

	private volatile byte[] internalError;
	private byte[] getInternalErrorResult() {
		if (this.internalError == null) {
			this.internalError = AmfUtil.encode(
					new Result(null, Result.FATAL_INTERNAL, "Server Internal Error."));
		}
		return this.internalError;
	}
}
